==> adeptutils-0.0.2.tar.gz: lib/src/extendables/adept_state.dart <==
import 'package:flutter/material.dart';
import 'dart:async';
import 'package:adeptutils/src/interfaces/iwidget.dart';
import 'package:adeptutils/src/collections/stack.dart';
import 'package:modal_progress_hud/modal_progress_hud.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:flutter/scheduler.dart';
import 'package:adeptutils/src/interfaces/ibase.dart';
import 'package:adeptutils/src/exception_utils.dart';
import 'package:adeptutils/src/enumerators/lifecycle_state.dart';

abstract class AdeptState<W extends StatefulWidget, T> extends State<W> with WidgetsBindingObserver implements IWidget<T>{
  final _scaffoldKey = GlobalKey<ScaffoldState>();
  GlobalKey<ScaffoldState> get scaffoldKey => _scaffoldKey;
  BuildContext _context;
  BuildContext get context => _context;
  IBase _base;
  IBase get base => _base;
  int _exitControl = 0;
  String _loadingMessage;
  Widget _progressIndicator;
  bool _isLoading = false, _showProgressIndeterminate = true, _pressBackAgain = false, _isBuildCompleted = false;
  get isBuildCompleted => _isBuildCompleted;
  String _pressBackAgainMessage;
  Widget fragment;
  ListStack<Widget> fragmentsManager = new ListStack();
  LifecycleState _lifecycleState = LifecycleState.created;
  LifecycleState get lifecycleState => _lifecycleState;
  Set<StreamSubscription> _streamSubscriptions = new Set();

  AdeptState({IBase base, bool pressBackAgain = false, String pressBackAgainMessage = "Press back again to exit"}) :
  _base = base,
  _pressBackAgain = pressBackAgain,
  _pressBackAgainMessage = pressBackAgainMessage
  ;

  void handleException(dynamic e,
      {
        BuildContext context, String title = "Error",
        String btnContinueText = "CONTINUE",
        String btnExitText = "EXIT"
      }) {
    if(context == null) context = this.context;
    ExceptionUtils.handleException(e, context, title: title,
        btnContinueText: btnContinueText, btnExitText: btnExitText);
  }

  bool addLifecycleAware(StreamSubscription streamSubscription) =>
      streamSubscription != null ? _streamSubscriptions.add(streamSubscription) : false;

  bool removeLifecycleAware(StreamSubscription streamSubscription) =>
      streamSubscription != null ? _streamSubscriptions.remove(streamSubscription) : false;

  void _notifyLifecycleChange(LifecycleState lifecycleState){
    if(lifecycleState == null) return;
    if(lifecycleState != LifecycleState.resumed &&
        lifecycleState != LifecycleState.paused &&
        lifecycleState != LifecycleState.stopped &&
        lifecycleState != LifecycleState.destroyed) return;
    _streamSubscriptions.forEach((streamSubscription) {
      switch(lifecycleState){
        case LifecycleState.resumed:
          streamSubscription.resume();
          break;
        case LifecycleState.paused:
        case LifecycleState.stopped:
          streamSubscription.pause();
        break;
        case LifecycleState.destroyed:
          streamSubscription.cancel();
          break;
        default:
      }
    });
  }

  @override
  void initState() {
    super.initState();
    WidgetsBinding.instance.addObserver(this);
    SchedulerBinding.instance.addPostFrameCallback((_) {
      _notifyLifecycleChange(_lifecycleState = LifecycleState.created);
      onCreated();
    });
  }

  @override
  void dispose() {
    WidgetsBinding.instance.removeObserver(this);
    _lifecycleState = LifecycleState.destroyed;
    _notifyLifecycleChange(lifecycleState);
    onDestroy();
    super.dispose();
  }

  @override
  void deactivate() {
    if(_lifecycleState == LifecycleState.paused){
      _lifecycleState = LifecycleState.resumed;
      _notifyLifecycleChange(lifecycleState);
      onResumed();
    } else{
      _lifecycleState = LifecycleState.paused;
      _notifyLifecycleChange(lifecycleState);
      onPaused();
    }
    super.deactivate();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if(state == null) return;
    switch(state){
      case AppLifecycleState.resumed:
        _lifecycleState = LifecycleState.resumed;
        _notifyLifecycleChange(lifecycleState);
        onResumed();
        break;
      case AppLifecycleState.inactive:
        _lifecycleState = LifecycleState.inactive;
        _notifyLifecycleChange(lifecycleState);
        onInactive();
        break;
      case AppLifecycleState.paused://AppLifecycleState.paused is used as stopped state because deactivate() works more as a pause for lifecycle
        _lifecycleState = LifecycleState.stopped;
        _notifyLifecycleChange(lifecycleState);
        onStopped();
        break;
      case AppLifecycleState.suspending:
        _lifecycleState = LifecycleState.suspending;
        _notifyLifecycleChange(lifecycleState);
        onSuspending();
        break;
    }
  }

  void refresh([VoidCallback fn]) {
    setState(fn ?? (){});
  }

  @mustCallSuper
  @override
  Widget build(BuildContext context) {
    _context = context;
    _isBuildCompleted = false;
    SchedulerBinding.instance.addPostFrameCallback((_) => onBuildCompleted());
    return null;
  }

  bool get isLoading => _isLoading;

  Widget _buildProgressIndicator() =>
    _progressIndicator ??
    Column(
      mainAxisAlignment: MainAxisAlignment.center,
      mainAxisSize: MainAxisSize.min,
      children: <Widget>[
        _showProgressIndeterminate ? CircularProgressIndicator() : Container(),
        Divider(color: Colors.transparent,),
        _loadingMessage != null && _loadingMessage.isNotEmpty ?
          Text(
            _loadingMessage,
            textAlign: TextAlign.center,
            style: TextStyle(color: Colors.white),
          ) : Container(),
      ],
    );

  Widget bodyLoading({
    @required final Widget body,
    final bool inAsyncCall,
    final double opacity = 0.7,
    final Color color = Colors.black,
    final Widget progressIndicator,
    final Offset offset,
    final bool dismissible = false,
  }) =>
      ModalProgressHUD(
        child: body,
        inAsyncCall: inAsyncCall ?? _isLoading,
        opacity: opacity,
        color: color,
        progressIndicator: progressIndicator ?? _buildProgressIndicator(),
        offset: offset,
        dismissible: dismissible,
      );

  WillPopScope bodyOnBackPressed({
    @required final Widget body,
  }) =>
      WillPopScope(
        child: body,
        onWillPop: onBackPressed,
      );

  Widget setFragment(Widget fragment)
  {
    if(fragmentsManager.isNotEmpty) fragmentsManager.pop();
    return addFragment(fragment);
  }

  Widget addFragment(Widget fragment, {bool repeated = false}) {
    if(!repeated && fragmentsManager.isNotEmpty) {
      Widget currentFragment = fragmentsManager.peek();
      if(currentFragment != null && fragment != null){
        if(currentFragment.runtimeType == fragment.runtimeType)
          return fragment;
      }
    }
    this.fragment = fragmentsManager.push(fragment);
    setState(() {});
    return fragment;
  }

  void clearFragments() {
    this.fragment = null;
    fragmentsManager.clear();
  }

  bool isCurrentFragment(Widget fragment) => fragmentsManager.isNotEmpty ? fragmentsManager.peek() == fragment : false;

  ScaffoldFeatureController<SnackBar, SnackBarClosedReason> showSnackBar(SnackBar snackbar, {GlobalKey<ScaffoldState> scaffoldKey}) {
    if(snackbar != null) {
      scaffoldKey ??= this.scaffoldKey;
      return scaffoldKey.currentState.showSnackBar(snackbar);
    }
    return null;
  }

  @override
  void showLoading({String message, bool showProgressIndeterminate = true, Widget progressIndicator}) {
    _loadingMessage = message;
    _showProgressIndeterminate = showProgressIndeterminate ?? true;
    this._progressIndicator = progressIndicator;
    setState(() => _isLoading = true);
  }

  @override
  void hideLoading({int delayMillis}) async{
    if(delayMillis != null)
      return Future.delayed(Duration(milliseconds: delayMillis), () => hideLoading());
    setState(() => _isLoading = false);
  }

  @override
  void extras() {

  }

  @override
  void init() {

  }

  @override
  void check() {

  }

  @override
  void listeners() {

  }

  @override
  bool validate() => true;

  @override
  int getSelectedDrawerItemId() => null;

  @override
  bool isPressBackAgain() => _pressBackAgain;

  @override
  void pressBackAgain(bool pressBackAgain, {String pressBackAgainMessage = "Press back again to exit"}) {
    _pressBackAgain = pressBackAgain;
    _pressBackAgainMessage = pressBackAgainMessage;
  }

  @override
  void onBuildCompleted() {
    _isBuildCompleted = true;
  }


  /// onBackPressed returns false to stop propagating the event, true in other cases
  @override
  Future<bool> onBackPressed() async {
    if(fragmentsManager.size > 1){
      fragmentsManager.pop();
      fragment = fragmentsManager.peek();
      setState(() {});
      return false;
    }
    if(isPressBackAgain())
    {
      if((DateTime.now().millisecondsSinceEpoch - _exitControl) <= 3000)
        return true;
      else
      {
        _exitControl = DateTime.now().millisecondsSinceEpoch;
        Fluttertoast.showToast(msg: _pressBackAgainMessage, toastLength: Toast.LENGTH_LONG);
        return false;
      }
    }
    return true;
  }

  @override
  Future<T> push<T>({BuildContext context, Route<T> route, Widget widget}) {
    if(context == null) context = _context;
    if(route != null) return Navigator.push(context, route); else
    if(widget != null) return Navigator.push(context, MaterialPageRoute(builder: (context) => widget));
    return null;
  }

  @override
  Future<T> pushReplacement<T, TO>({BuildContext context, Route<T> newRoute, Widget newWidget, TO result}) {
    if(context == null) context = _context;
    if(newRoute != null) return Navigator.pushReplacement(context, newRoute); else
    if(newWidget != null) return Navigator.pushReplacement(context, MaterialPageRoute(builder: (context) => newWidget));
    return null;
  }

  @override
  void onClick({Widget w, id}) {

  }

  @override
  void onLongClick({Widget w, id}) {

  }

  @override
  void onItemLongClick({Widget w, T item, int position, id}) {

  }

  @override
  void onItemClick({Widget w, T item, int position, id}) {

  }

  @override
  void onCreated() {
//    print("onCreated");
  }

  @override
  void onResumed() {
//    print("onResumed");
  }

  @override
  void onInactive() {
//    print("onInactive");
  }

  @override
  void onPaused() {
//    print("onPaused");
  }

  @override
  void onSuspending() {
//    print("onSuspending");
  }

  @override
  void onStopped() {
//    print("onStopped");
  }

  @override
  void onDestroy() {
//    print("onDestroy");
  }

}

==> adeptutils-1.0.0.tar.gz: lib/src/extendables/adept_state.dart <==
import 'package:flutter/material.dart';
import 'dart:async';
import 'package:adeptutils/src/interfaces/iwidget.dart';
import 'package:adeptutils/src/collections/stack.dart';
import 'package:modal_progress_hud/modal_progress_hud.dart';
import 'package:fluttertoast/fluttertoast.dart';
import 'package:flutter/scheduler.dart';
import 'package:adeptutils/src/interfaces/ibase.dart';
import 'package:adeptutils/src/exception_utils.dart';
import 'package:adeptutils/src/enumerators/lifecycle_state.dart';

abstract class AdeptState<W extends StatefulWidget, T> extends State<W> with WidgetsBindingObserver implements IWidget<T>{
  final _scaffoldKey = GlobalKey<ScaffoldState>();
  GlobalKey<ScaffoldState> get scaffoldKey => _scaffoldKey;
  BuildContext _context;
  BuildContext get context => _context;
  IBase _base;
  IBase get base => _base;
  int _exitControl = 0;
  String _loadingMessage;
  Widget _progressIndicator;
  bool _isLoading = false, _showProgressIndeterminate = true, _pressBackAgain = false, _isBuildCompleted = false;
  get isBuildCompleted => _isBuildCompleted;
  String _pressBackAgainMessage;
  Widget fragment;
  ListStack<Widget> fragmentsManager = new ListStack();
  LifecycleState _lifecycleState = LifecycleState.created;
  LifecycleState get lifecycleState => _lifecycleState;
  Set<StreamSubscriptionLifecycleAware> _streamSubscriptionsLifecycleAware = new Set();

  AdeptState({IBase base, bool pressBackAgain = false, String pressBackAgainMessage = "Press back again to exit"}) :
  _base = base,
  _pressBackAgain = pressBackAgain,
  _pressBackAgainMessage = pressBackAgainMessage
  ;

  void handleException(dynamic e,
      {
        BuildContext context, String title = "Error",
        String btnContinueText = "CONTINUE",
        String btnExitText = "EXIT"
      }) {
    if(context == null) context = this.context;
    ExceptionUtils.handleException(e, context, title: title,
        btnContinueText: btnContinueText, btnExitText: btnExitText);
  }

  ///The streamSubscription to be aware of life cycle changes
  ///The lifecycleStates to be considered during changes, by default null means all lifeCycles
  bool addLifecycleAware(StreamSubscription streamSubscription, {Set<LifecycleState> lifecycleStates}) =>
      streamSubscription != null ? _streamSubscriptionsLifecycleAware.add(StreamSubscriptionLifecycleAware(streamSubscription: streamSubscription, lifecycleStates: lifecycleStates)) : false;

  bool removeLifecycleAware(StreamSubscription streamSubscription) {
    if(streamSubscription == null) return false;
    bool removed = false;
    _streamSubscriptionsLifecycleAware.removeWhere((streamSubscriptionLifecycleAware) =>
      streamSubscriptionLifecycleAware?.streamSubscription == streamSubscription
    );
    return removed;
  }

  void _notifyLifecycleChange(LifecycleState lifecycleState){
    if(lifecycleState == null) return;
    if(lifecycleState != LifecycleState.resumed &&
        lifecycleState != LifecycleState.paused &&
        lifecycleState != LifecycleState.stopped &&
        lifecycleState != LifecycleState.destroyed) return;
    _streamSubscriptionsLifecycleAware.forEach((streamSubscriptionLifecycleAware) =>
      streamSubscriptionLifecycleAware?.onLifecycleChanged(lifecycleState)
    );
  }

  @override
  void initState() {
    super.initState();
    WidgetsBinding.instance.addObserver(this);
    SchedulerBinding.instance.addPostFrameCallback((_) {
      _notifyLifecycleChange(_lifecycleState = LifecycleState.created);
      onCreated();
    });
  }

  @override
  void dispose() {
    WidgetsBinding.instance.removeObserver(this);
    _lifecycleState = LifecycleState.destroyed;
    _notifyLifecycleChange(lifecycleState);
    onDestroy();
    super.dispose();
  }

  @override
  void deactivate() {
    if(_lifecycleState == LifecycleState.paused){
      _lifecycleState = LifecycleState.resumed;
      _notifyLifecycleChange(lifecycleState);
      onResumed();
    } else{
      _lifecycleState = LifecycleState.paused;
      _notifyLifecycleChange(lifecycleState);
      onPaused();
    }
    super.deactivate();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if(state == null) return;
    switch(state){
      case AppLifecycleState.resumed:
        _lifecycleState = LifecycleState.resumed;
        _notifyLifecycleChange(lifecycleState);
        onResumed();
        break;
      case AppLifecycleState.inactive:
        _lifecycleState = LifecycleState.inactive;
        _notifyLifecycleChange(lifecycleState);
        onInactive();
        break;
      case AppLifecycleState.paused://AppLifecycleState.paused is used as stopped state because deactivate() works more as a pause for lifecycle
        _lifecycleState = LifecycleState.stopped;
        _notifyLifecycleChange(lifecycleState);
        onStopped();
        break;
      case AppLifecycleState.suspending:
        _lifecycleState = LifecycleState.suspending;
        _notifyLifecycleChange(lifecycleState);
        onSuspending();
        break;
    }
  }

  void refresh([VoidCallback fn]) {
    setState(fn ?? (){});
  }

  @mustCallSuper
  @override
  Widget build(BuildContext context) {
    _context = context;
    _isBuildCompleted = false;
    SchedulerBinding.instance.addPostFrameCallback((_) => onBuildCompleted());
    return null;
  }

  bool get isLoading => _isLoading;

  Widget _buildProgressIndicator() =>
    _progressIndicator ??
    Column(
      mainAxisAlignment: MainAxisAlignment.center,
      mainAxisSize: MainAxisSize.min,
      children: <Widget>[
        _showProgressIndeterminate ? CircularProgressIndicator() : Container(),
        Divider(color: Colors.transparent,),
        _loadingMessage != null && _loadingMessage.isNotEmpty ?
          Text(
            _loadingMessage,
            textAlign: TextAlign.center,
            style: TextStyle(color: Colors.white),
          ) : Container(),
      ],
    );

  Widget bodyLoading({
    @required final Widget body,
    final bool inAsyncCall,
    final double opacity = 0.7,
    final Color color = Colors.black,
    final Widget progressIndicator,
    final Offset offset,
    final bool dismissible = false,
  }) =>
      ModalProgressHUD(
        child: body,
        inAsyncCall: inAsyncCall ?? _isLoading,
        opacity: opacity,
        color: color,
        progressIndicator: progressIndicator ?? _buildProgressIndicator(),
        offset: offset,
        dismissible: dismissible,
      );

  WillPopScope bodyOnBackPressed({
    @required final Widget body,
  }) =>
      WillPopScope(
        child: body,
        onWillPop: onBackPressed,
      );

  Widget setFragment(Widget fragment)
  {
    if(fragmentsManager.isNotEmpty) fragmentsManager.pop();
    return addFragment(fragment);
  }

  Widget addFragment(Widget fragment, {bool repeated = false}) {
    if(!repeated && fragmentsManager.isNotEmpty) {
      Widget currentFragment = fragmentsManager.peek();
      if(currentFragment != null && fragment != null){
        if(currentFragment.runtimeType == fragment.runtimeType)
          return fragment;
      }
    }
    this.fragment = fragmentsManager.push(fragment);
    setState(() {});
    return fragment;
  }

  void clearFragments() {
    this.fragment = null;
    fragmentsManager.clear();
  }

  bool isCurrentFragment(Widget fragment) => fragmentsManager.isNotEmpty ? fragmentsManager.peek() == fragment : false;

  ScaffoldFeatureController<SnackBar, SnackBarClosedReason> showSnackBar(SnackBar snackbar, {GlobalKey<ScaffoldState> scaffoldKey}) {
    if(snackbar != null) {
      scaffoldKey ??= this.scaffoldKey;
      return scaffoldKey.currentState.showSnackBar(snackbar);
    }
    return null;
  }

  @override
  void showLoading({String message, bool showProgressIndeterminate = true, Widget progressIndicator}) {
    _loadingMessage = message;
    _showProgressIndeterminate = showProgressIndeterminate ?? true;
    this._progressIndicator = progressIndicator;
    setState(() => _isLoading = true);
  }

  @override
  void hideLoading({int delayMillis}) async{
    if(delayMillis != null)
      return Future.delayed(Duration(milliseconds: delayMillis), () => hideLoading());
    setState(() => _isLoading = false);
  }

  @override
  void extras() {

  }

  @override
  void init() {

  }

  @override
  void check() {

  }

  @override
  void listeners() {

  }

  @override
  bool validate() => true;

  @override
  int getSelectedDrawerItemId() => null;

  @override
  bool isPressBackAgain() => _pressBackAgain;

  @override
  void pressBackAgain(bool pressBackAgain, {String pressBackAgainMessage = "Press back again to exit"}) {
    _pressBackAgain = pressBackAgain;
    _pressBackAgainMessage = pressBackAgainMessage;
  }

  @override
  void onBuildCompleted() {
    _isBuildCompleted = true;
  }

  /// onBackPressed returns false to stop propagating the event, true in other cases
  @override
  Future<bool> onBackPressed() async {
    if(fragmentsManager.size > 1){
      fragmentsManager.pop();
      fragment = fragmentsManager.peek();
      setState(() {});
      return false;
    }
    if(isPressBackAgain())
    {
      if((DateTime.now().millisecondsSinceEpoch - _exitControl) <= 3000)
        return true;
      else
      {
        _exitControl = DateTime.now().millisecondsSinceEpoch;
        Fluttertoast.showToast(msg: _pressBackAgainMessage, toastLength: Toast.LENGTH_LONG);
        return false;
      }
    }
    return true;
  }

  @override
  Future<T> push<T>({BuildContext context, Route<T> route, String routeName, Widget widget}) {
    if(context == null) context = _context;
    if(route != null) return Navigator.push(context, route); else
    if(widget != null) return Navigator.push(context, MaterialPageRoute(builder: (context) => widget, settings: routeName != null ? RouteSettings(name: routeName) : null));
    return null;
  }

  @override
  Future<T> pushReplacement<T, TO>({BuildContext context, Route<T> newRoute, String newRouteName, Widget newWidget, TO result}) {
    if(context == null) context = _context;
    if(newRoute != null) return Navigator.pushReplacement(context, newRoute); else
    if(newWidget != null) return Navigator.pushReplacement(context, MaterialPageRoute(builder: (context) => newWidget, settings: newRouteName != null ? RouteSettings(name: newRouteName) : null));
    return null;
  }

  @override
  void onClick({Widget w, id}) {

  }

  @override
  void onLongClick({Widget w, id}) {

  }

  @override
  void onItemLongClick({Widget w, T item, int position, id}) {

  }

  @override
  void onItemClick({Widget w, T item, int position, id}) {

  }

  @override
  void onCreated() {
//    print("onCreated");
  }

  @override
  void onResumed() {
//    print("onResumed");
  }

  @override
  void onInactive() {
//    print("onInactive");
  }

  @override
  void onPaused() {
//    print("onPaused");
  }

  @override
  void onSuspending() {
//    print("onSuspending");
  }

  @override
  void onStopped() {
//    print("onStopped");
  }

  @override
  void onDestroy() {
//    print("onDestroy");
  }

}
